/** This code is under the GPL, viewable at http://www.gnu.org/copyleft/gpl.html
 *  You are free to modify and redistribute the code as you see fit, but three main 
 *  conditions exist for usage of this code:
 *  1.  You may only pass on this code with the same rights and priveleges that were given
 *      to you (In other words, you must redistribute this license)
 *  2.  You may NOT distribute this code for sale or for profit.
 *  3.  There are no warranties about this code.  It is designed not to screw up your computer,
 *      but in the case that it does, the writer cannot be held responsible for losses.
 *
 *  Code author:  flug_m (post comments/questions to http://www.saveourtunes.com/forum/ In future
 *                versions, I'll make available alternative means of contact.)
 *  Code Revision date:  September 1, 2007
 *  Code Revision version: 0.5
 *
 *  As its name implies, all the objects stored in this class should be shared
 *  between all the classes, particularly the threads, of this program.
 *
 *  Any object registered in this class should provide a means that make
 *  it accessible to all classes in the program.
 *
 *  Changes:
 *   8/30/07  Added support for saving ourTunes configuration.
 *   8/30/07  The ourTunes frame registers in the shared.  This permits updating of the status pane.
 *   8/30/07  Networking classes will throw errors if no connected NIC exists.  ourTunes displays
 *            this error message and closes out.
 *   8/31/07  Added support for the ourtunes.os package.  Essentially, you must be using Windows or
 *            Mac. (For all Linux users, implement the ourtunes.os.OS interface and leave me a note
 *            asking for its inclusion).
 *   8/31/07  Changed the way sequential numbers are generated; they interact with the config object.
 *   9/06/07  the variable server switched from a string to an InetAddress.
 *  11/16/07  MP3 tagging support added.
 *  11/17/07  M4A tagging support added.
 *  12/05/07  Auto-generated filenames are no longer illegal in Windows XP or greater (assumption is
 *				made that the filenames will be sanitized under Mac OS X also).
 *  1/11/08  Discovered Socket execption error in the static block.  Rewrote an error message.
 *  3/21/08  Due to changes between v1 and v2 of JmDNS library, code that instantiated the library
 *           was re-written to correctly call methods out of the library.
 *
 *  7/23/08  Support added for midi and wav files (generic song files).
 *  8/24/08  Fix problems with String.replaceAll() - eseentially needed to replace each char with the regex pattern string.
 */
 package ourtunes;
 
import java.util.*;
import java.text.*;
import java.io.File;
import javax.jmdns.*;
import java.net.*;
import javax.swing.JOptionPane;

import ourtunes.os.*;
import ourtunes.gui.OurTunesPanel;
import ourtunes.music.*;

public class Shared
{
	static
	{
		String host = null;
		try{
		host = InetAddress.getLocalHost().toString();
		System.out.println(InetAddress.getLocalHost());
		bonjour = JmDNS.create(InetAddress.getLocalHost());
		fmt = new DecimalFormat("0000");
		
		} catch (SocketException e)
		{
			JOptionPane.showMessageDialog(null, "ourtunes could not open a socket on the "
				+"\nnetwork interface with this IP address:  "+(host==null?"none":host)+"."
				+" Check your firewall and connection and try again.");
			e.printStackTrace();
			System.exit(1);
		}
		catch (Exception e)
		{
			System.out.println("could not start ourTunes:  ");
			e.printStackTrace();
			System.exit(1);
		}
		
		//Provide support for the configuration of ourTunes.
		String os = System.getProperty("os.name").toLowerCase();
		if (os.indexOf("windows") != -1)
			opSys = new Windows();
		else if (os.indexOf("mac") != -1)
			opSys = new Mac(); 
		else
		{
			JOptionPane.showMessageDialog(null, "You are not using a mac/windows computer.  "+
				"\nYou'll need iTunes to use ourTunes.  This program will exit.\n"+
				"Detected Operating System:  "+System.getProperty("os.name"));
			System.exit(0);
		}
	}
	
	/**
	 *Generate a random number, and format as a 4 digits (use leading 0s)
	 *Return as a formatted String.
	 */
	public static String getRandom()
	{
		int x = (int) (Math.random() * 10000);
		return fmt.format(x);
		
	}
	/**
	 *Generate a sequential number, based off of program settings
	 *Return a formatted String.
	 */
	public static String getSequential()
	{
		int count = Integer.parseInt(config.getProperty("counter"));
		config.setProperty("counter", Integer.toString(++count));
		return fmt.format(count);
	}
	/**
	 *Convenience method to format any number into the common 4-digit string.
	 *This method will modulus by 10,000 to get 4 digits.
	 *@Returns:  the appropriately formatted string.
	 */
	public static String formatAny(int x)
	{
		return fmt.format(x%10000);
	}
	
	/**
	 *fix the file name.  remove the illegal characters from the expression,
	 *and place it into a specified directory.  Have yet to come up with that,
	 *so it goes into the current directory.
	 *
	 *A purposely placed this method here so that it can be quickly expanded
	 *to place a file into any directory.  This method should also
	 *provide support for user defined file names (not supported in build a).
	 *
	 *While I'm at it, I need all the illegal characters for all Operating Systems,
	 *and those need to be removed.  I haven't tested the current implementation
	 *to remove those illegal characters.
	 */
	public static void fixFile(File f, String extension)
	{
		
		Song s = null;
		String name = panel.getFmtText();
		//name = name.replaceAll("[/?<>\\:*|^\"]", "_");
		try{
		if (extension.equalsIgnoreCase("mp3"))
		{
			s = new ID3Tag(f);
		}
		else if (extension.equalsIgnoreCase("m4a"))
		{
			s = new AAC(f);
		}
		else if (extension.equalsIgnoreCase("midi") || extension.equalsIgnoreCase("mid") || 
			     extension.equalsIgnoreCase("wav"))
		
		{
			String fname = 
			JOptionPane.showInputDialog(panel, "What is the name of the "+extension+" file?", "File Name...");
			s = new GenericSong(f,fname);	
		}	
		else {
			System.out.println("Unknown file:  "+extension);
			name = getRandom()+extension;
		}
			if(name.indexOf("%R") != -1) name = name.replaceAll("%R", getSequential());
			name = name.replaceAll("%r", getRandom());
			name = name.replaceAll("%e", extension);
			name = name.replaceAll("%a", s.getArtist());
			name = name.replaceAll("%t", s.getTitle());
			if (s instanceof GenericSong)
				name = name.concat(s.toString());
			System.out.println(name);
			//If the filename contains illegal characters, replace them
			name = name.replaceAll("\\\\","_");
			name = name.replaceAll("/","_");
			name = name.replaceAll("<","_");
			name = name.replaceAll(">","_");
			name = name.replaceAll("\"","_");
			name = name.replaceAll("\\?","_");
			name = name.replaceAll("\\|","_");
			name = name.replaceAll("\\*","_");
			name = name.replaceAll(":","_");
			name = name.replaceAll("%","_");
			
			s.close();			
		} catch (java.io.IOException e) {name = getRandom()+"."+extension;}
		
		
		f.renameTo(new File(getProperty("home"), name));
	}
	
	/**
	 * Determine if an InetAddress belongs to a NetworkInterface
	 * on this computer.  Using the loopback address should return true.
	 *
	 * @Returns:  true if any NIC owns the InetAddress, false otherwise. 
	 */
	public static boolean isHostAddress(InetAddress addr)
	{
		try{
		NetworkInterface nic = NetworkInterface.getByInetAddress(addr);
		return (nic != null);
		} catch(Exception e) {return false;}
	}
	/**
	 *Set the server address for the pass-through thread.
	 *This is data that needs to be passed from a GUI thread
	 *to a networking thread, and this is the easist way to do that.
	 */
	public static void setServer(InetAddress svr)
	{
		System.out.println("setServer:  "+svr);
		server = svr;
	}
	/**Obtain the server address registered
	 * last registered with this class.  If no address
	 * has been registered, this method will return null.
	 */
	public static InetAddress getServer()
	{
		return server;
	}
	/**
	 * Conveience method for formatting a ServiceInfo's data
	 * into HTML (all swing components except the JTextArea support
	 * HTML 3.2 + CSS1).
	 *
	 * This method has a very narrow purpose.
	 */
	public static String getHTMLText(ServiceInfo info)
	{
		//Note, this method just inserts "<br>" instead of line breaks for display purposes.
		StringBuffer buf = new StringBuffer();
		
		//code copy: com.strangeberry.jmdns.tools.Browser
		for (Enumeration names = info.getPropertyNames() ; names.hasMoreElements() ; ) {
                    String prop = (String)names.nextElement();
                    buf.append(prop);
                    buf.append('=');
                    buf.append(info.getPropertyString(prop));
                    buf.append("<br>");
                }
         return buf.toString();       
	}
		
	/**
	 *Get the shared JmDNS object for the program.
	 *This is designed to prevent multiple mDNS objects
	 *from running.
	 */
	public static JmDNS getJmDNSInstance()
	{
		return bonjour;
	}
	
	/**
	 *Get the OS for registered in this Shared class.  The OS
	 *will list the default directory, but the Configuration file
	 *will list the directory the user wants the music to be downloaded to.
	 */
	public static OS getOS()
	{
		return opSys;
	}
	
	/**
	 *set the properties object.  The properties object manages the configuration of
	 *ourTunes.
	 */
	 public static void setConfiguration(Properties p)
	 {
	 	config = p;
	 }
	 /**
	  *get a property from the configuration settings
	  *
	  *@Returns:  The associated property of the given key, null if no value exists.
	  */
	  public static String getProperty(String key)
	  {
	  		return config.getProperty(key);
	  }
	 /**
	  *set a property in the configuration settings.
	  */
	  public static void setProperty(String key, String value)
	  {
	  	config.setProperty(key, value);
	  }
	
	//Convenience methods for adding Services/ServiceListeners.
	public static void addServiceListener(String service, ServiceListener listener)
	{
		bonjour.addServiceListener(service, listener);
	}
	public static void registerService(ServiceInfo info) throws Exception
	{
		bonjour.registerService(info);
	}
	
	//class objects
	private static NumberFormat fmt;
	private static JmDNS bonjour;
	private static InetAddress server;
	private static OS opSys;
	
	protected static Properties config;	
	public static OurTunesPanel panel;
}