/** This code is under the GPL, viewable at http://www.gnu.org/copyleft/gpl.html
 *  You are free to modify and redistribute the code as you see fit, but three main 
 *  conditions exist for usage of this code:
 *  1.  You may only pass on this code with the same rights and priveleges that were given
 *      to you (In other words, you must redistribute this license)
 *  2.  You may NOT distribute this code for sale or for profit.
 *  3.  There are no warranties about this code.  It is designed not to screw up your computer,
 *      but in the case that it does, the writer cannot be held responsible for losses.
 *
 *  Code author:  secret_squirrel (member:http://www.saveourtunes.com/)
 *  Code Revision date:  November 15, 2007
 *  Code Revision version: 0.5
 *
 *  MP3 wrapper class to represent/extract data from MP3 files.
 *
 *  Changes:
 *  11/15/2007 @author flug_m  class implements Song interface
 *  11/15/2007 @author flug_m  addition of finalize() method to free system resources when class is not needed. 
 *  11/15/2007 @author flug_m  placed into ourtunes.music.* package
 *  11/16/2007 @author flug_m  changed signature to method getYear() (initial signature:  public String getYear())
 *  2/22/2008  @author flug_m  many get() methods have try...catch blocks so that they give back a blank
 *                             record instead of throwing an exeception.
 */
package ourtunes.music;

import java.io.*;
import java.util.*;
import java.util.zip.Adler32;
import java.util.Date;

/**
 * A class for extracting the ID3 tag from an MP3 file. Usage:
 * ID3Tag tag = new ID3Tag(file);
 * tag.close();
 * String name = tag.getName();
 *
 */
public class ID3Tag implements ourtunes.music.Song
{
    private static final String[] genres = {
        "Blues", "Classic Rock", "Country", "Dance", "Disco", "Funk", "Grunge", "Hip-Hop",
        "Jazz", "Metal", "New Age", "Oldies", "Other", "Pop", "R&B", "Rap", "Reggae",
        "Rock", "Techno", "Industrial", "Alternative", "Ska", "Death Metal", "Pranks",
        "Soundtrack", "Euro-Techno", "Ambient", "Trip-Hop", "Vocal", "Jazz+Funk", "Fusion",
        "Trance", "Classical", "Instrumental", "Acid", "House", "Game", "Sound Clip", "Gospel",
        "Noise", "AlternRock", "Bass", "Soul", "Punk", "Space", "Meditative", "Instrumental Pop",
        "Instrumental Rock", "Ethnic", "Gothic", "Darkwave", "Techno-Industrial", "Electronic",
        "Pop-Folk", "Eurodance", "Dream", "Southern Rock", "Comedy", "Cult", "Gangsta", "Top 40",
        "Christian Rap", "Pop/Funk", "Jungle", "Native American", "Cabaret", "New Wave",
        "Psychadelic", "Rave", "Showtunes", "Trailer", "Lo-Fi", "Tribal", "Acid Punk",
        "Acid Jazz", "Polka", "Retro", "Musical", "Rock & Roll", "Hard Rock", "Folk",
        "Folk-Rock", "National Folk", "Swing", "Fast Fusion", "Bebob", "Latin", "Revival",
        "Celtic", "Bluegrass", "Avantgarde", "Gothic Rock", "Progressive Rock",
        "Psychedelic Rock", "Symphonic Rock", "Slow Rock", "Big Band", "Chorus", "Easy Listening",
        "Acoustic", "Humour", "Speech", "Chanson", "Opera", "Chamber Music", "Sonata", "Symphony",
        "Booty Bass", "Primus", "Porn Groove", "Satire", "Slow Jam", "Club", "Tango", "Samba",
        "Folklore", "Ballad", "Power Ballad", "Rhythmic Soul", "Freestyle", "Duet", "Punk Rock",
        "Drum Solo", "A capella", "Euro-House", "Dance Hall"
     };

    private HashMap tags;
    private boolean vbr, hasid3v1;
    private int id3version;
    private int bitrate, samplerate, time, trackno, mode, id3v2end;
    private float gain;
    private String origgain;
    private long size, sum;
    private Date date;
    private RandomAccessFile rfile;

    private static final int V1=3, V2=2, V25=0;
    private static final int MODE_JOINTSTEREO=1, MODE_STEREO=2, MODE_MONO=3;

    /**
     * Extracts an ID3 tag from the given file
     * @param file The file containing an ID3 tag
     */
    public ID3Tag(File file) throws IOException {
	rfile = null;
        size = file.length();
        date = new Date(file.lastModified());
        tags = new HashMap();
	try {
	    rfile = new RandomAccessFile(file, "r");
	    id3version = readID3v2(rfile, file.toString(), tags);
            if (id3version<0) {
                id3version=0;
                if (readID3v1(rfile, tags)) {
                    id3version = 0x010000;
                    hasid3v1 = true;
                } else {
                    guessDetails(file, tags);
                }
            }
	    readBitrate(file.toString(), rfile);
	} catch (IOException e) {
            try {
                if (rfile!=null) rfile.close();
            } catch (Exception e2) {}
	    throw e;
	}

        for (Iterator i = tags.entrySet().iterator();i.hasNext();) {
            Map.Entry e = (Map.Entry)i.next();
            String key = (String)e.getKey();
            if (e.getValue() instanceof String) {
                String val = (String)e.getValue();
                if (key.equals("TRCK") || key.equals("TPOS")) {
                    if (val.indexOf("/")>=0) {
                        val = val.substring(0, val.indexOf("/"));
                        e.setValue(val);
                    }
                    if (val.equals("0")) val=null;
                }
                if (val==null || val.trim().length()==0) {
                    i.remove();
                } else {
                    if (val.indexOf(0)>=0) val=val.substring(val.indexOf(0)+1);

                    // This line may be commented out, but I found it handy to
                    // normalise the names first as iTunes on OS X stores the
                    // UTF-16 tags normalised in NFD form, which most browsers
                    // won't display properly. If all your ID3 tags are in
                    // ISO-8859-1 then it will make no difference either way.
                    //
                    //val = com.ibm.icu.text.Normalizer.normalize(val, com.ibm.icu.text.Normalizer.NFC);
                    //e.setValue(val.trim());
                }
            }
        }
        String genre = getTag("TCON");
        if (genre!=null && genre.charAt(0)=='(') {
            try {
                tags.put("TCON", genres[Integer.parseInt(genre.substring(1, genre.length()-1))]);
            } catch (Exception e) {
                tags.remove("TCON");
            }
        }
        tags.put("PATH", file.toString());
    }

    /**
     * Close the File the ID3Tag was reading from.
     */
    public void close() {
        try {
            if (rfile!=null) rfile.close();
        } catch (IOException e) {}
    }
	/**
	 * Finalize the object:  Make ready for garbage collection by the Java VM.
	 */
	 public void finalize()
	 {
	 	close();
	 }


    private int readSyncInt(RandomAccessFile in) throws IOException {
	return ((in.readByte()&0x7F)<<21) + ((in.readByte()&0x7F)<<14) + ((in.readByte()&0x7F)<<7) + (in.readByte()&0x7F);
    }

    private int readInt(RandomAccessFile in) throws IOException {
	return ((in.readByte()&0xFF)<<24) + ((in.readByte()&0xFF)<<16) + ((in.readByte()&0xFF)<<8) + (in.readByte()&0xFF);
    }

    /**
     * Guess tags from filename. Probably doesn't work very well
     */
    private void guessDetails(File file, HashMap tags) {
        int trackno=0, discno=0;
        String title=null, artist=null, album=null;
	String filename = file.getName().trim();
	int z = filename.lastIndexOf(".");
	if (z==filename.length()-4) filename=filename.substring(0,z);

	if (filename.length()>6 && Character.isDigit(filename.charAt(0)) && filename.charAt(1)=='-' && Character.isDigit(filename.charAt(2)) && (filename.charAt(4)==' ' || filename.charAt(4)=='.')) {
            discno = Integer.parseInt(filename.substring(0,1));
	    trackno = Integer.parseInt(filename.substring(2,2));
	    filename=filename.substring(5).trim();
        } else if (filename.length()>4 && Character.isDigit(filename.charAt(0)) && Character.isDigit(filename.charAt(1)) && (filename.charAt(2)==' ' || filename.charAt(2)=='.')) {
	    trackno = Integer.parseInt(filename.substring(0,2));
	    filename=filename.substring(3).trim();
	}

	z = filename.indexOf("-");
	if (z<0) {
	    title = filename;
	} else {
	    artist = filename.substring(0,z).trim();
	    title = filename.substring(z+1).trim();
	}

	file=file.getParentFile();
	if (file!=null) {
	    album = file.getName().trim();
	}

        if (discno!=0) tags.put("TPOS", ""+discno);
        if (trackno!=0) tags.put("TRCK", ""+trackno);
        if (title!=null) tags.put("TIT2", title);
        if (album!=null) tags.put("TALB", album);
        if (artist!=null) tags.put("TPE1", album);
    }

    /**
     * Read the ID3V1 tag. Only called if no ID3V2, so not tested for a while
     */
    private boolean readID3v1(RandomAccessFile in, HashMap tags) throws IOException {
        byte[] id3 = new byte[128];
	in.seek(in.length()-128);
	in.read(id3);
        if (id3[0]=='T' && id3[1]=='A' && id3[2]=='G') {
            tags.put("TIT2", new String(id3, 3, 30, "ISO-8859-1").trim());
            tags.put("TPE1", new String(id3, 33, 30, "ISO-8859-1").trim());
            tags.put("TALB", new String(id3, 63, 30, "ISO-8859-1").trim());
            tags.put("TYER", new String(id3, 93, 4, "ISO-8859-1").trim());
            tags.put("TRCK", ""+((int)id3[126]));
            tags.put("COMM", new String(id3, 97, (trackno==0 ? 30 : 28), "ISO-8859-1").trim());
            if (id3[127] < 126 && id3[127] >= 0) {
                tags.put("TCON", genres[id3[127]]);
            }
            return true;
	} else {
            return false;
        }
    }

    /**
     * Read the ID3V2 tag.
     */
    private int readID3v2(RandomAccessFile in, String filename, HashMap tags) throws IOException {
        boolean found=false;
        int id3version = -1;

	in.seek(0);
	while (in.readByte()=='I' && in.readByte()=='D' && in.readByte()=='3') {
            found=true;
	    int version = 0x020000 + (in.readByte()<<8) + in.readByte();
	    int id3flags = in.readByte();
	    id3v2end = (int)(readSyncInt(in)+in.getFilePointer());
//	    System.err.println("READ ID3 v"+Integer.toHexString(version)+" flags="+Integer.toHexString(id3flags)+" header to "+id3v2end);
	    if ((id3flags&0x40)==0x40) {
		int size = readInt(in);
		in.seek(in.getFilePointer()+size-4);
//		System.err.println("  SKIPPING EXTENDED HEADER OF "+size+" bytes");
	    }
	    if (version > id3version) { // Accept later versions over earlier
		id3version = version;
		if (version>=0x020300) {
		    do {
			byte[] b = new byte[4];
			in.read(b);
			String tag = new String(b, "ISO-8859-1");
			if (tag.equals("\0\0\0\0")) {
			    in.seek(id3v2end);
			} else {
                            int size = readInt(in);
			    int flags = in.read()<<8 + in.read();
//			    System.err.println("  READ TAG "+tag+" size="+size+" flags="+flags+" pointer="+in.getFilePointer());
			    if (size>id3v2end || size<0) {	// Catch bad data
				System.err.println("  BAD TAG "+tag+" in \""+filename+"\": skipping");
				in.seek(id3v2end);
			    } else {
				b = new byte[size];
				in.read(b);
				if (size>0 && (tag.charAt(0)=='T' || tag.charAt(0)=='W' || tag.equals("COMM") || tag.equals("IPLS") || tag.equals("XSOP") || tag.equals("USER"))) {
				    String encoding = b[0]==0 ? "ISO-8859-1" : "UTF-16";
                                    String lang = "";
				    try {
					String val;
                                        if ("COMM".equals(tag)) {
                                            val = new String(b, 4, b.length-4, encoding).trim();
                                            lang = new String(b, 1, 3, "ISO-8859-1");
                                            if (val.startsWith("iTunes_CDDB_1")) {
                                                tags.put("CDDB", val.substring(12));
                                            } else if (val.startsWith("iTunNORM\u0000")) {
                                                try {
                                                    StringTokenizer st = new StringTokenizer(val, " ");
                                                    st.nextToken();
                                                    float lval1 = Integer.parseInt(st.nextToken(), 16) / 1000f;
                                                    float rval1 = Integer.parseInt(st.nextToken(), 16) / 1000f;
                                                    float lval2 = Integer.parseInt(st.nextToken(), 16) / 2500f;
                                                    float rval2 = Integer.parseInt(st.nextToken(), 16) / 2500f;
                                                    if (lval2==0 && rval2==0) { lval2=lval1; rval2=rval1; }
                                                    gain = (float)(-10 * Math.log(lval2>rval2?lval2:rval2) / Math.log(10));
                                                    origgain = val.substring(10);
                                                    if (gain<-25 || gain>25 || gain!=gain || Float.isInfinite(gain)) {
                                                        gain = 0;
                                                        origgain = null;
                                                    }
                                                } catch (Exception e) { }
                                            } else if (!badComment(val)) {
                                                tags.put(tag, val);
                                            }
                                        } else if ("USER".equals(tag)) {
                                            val = new String(b, 4, b.length-4, encoding).trim();
                                            tags.put(tag, val);
                                        } else {
                                            val = new String(b, 1, b.length-1, encoding).trim();
                                            if (tag!=null) tags.put(tag, val);
                                        }
				    } catch (Exception e) { }
				} else if (tag.equals("APIC")) {
                                    setPictureTag(b, tags, false);
				} else if (tag.equals("MCDI")) {
                                    StringBuffer s = new StringBuffer(b.length*2);
                                    int len = b.length-1;
                                    while (len>0 && b[len]==0) len--;
                                    for (int i=0;i<len;i++) {
                                        int q = b[i]&0xFF;
                                        if (q<16) s.append("0");
                                        s.append(Integer.toHexString(q));
                                    }
                                    tags.put("MCDI", s.toString());
				} else if (tag.equals("UFID")) {
                                    if (b.length==12 && new String(b, 0, 7).equals("Adler32")) {
                                        this.sum = ((b[8]&0xFF)<<24) + ((b[9]&0xFF)<<16) + ((b[10]&0xFF)<<8) + ((b[11]&0xFF)<<0);
                                        tags.put("UFID", Long.toHexString(this.sum));
                                    }
                                }
			    }
			}
		    } while (in.getFilePointer()<id3v2end-10);
		} else {
		    do {
			byte[] b = new byte[3];
			in.read(b);
			String tag = new String(b, "ISO-8859-1").intern();
			if (tag.equals("\0\0\0")) {
			    in.seek(id3v2end);
			} else {
			    int size = (in.read()<<16) + (in.read()<<8) + in.read();
//			    System.err.println("  READ TAG "+tag+" size="+size);
			    if (size>id3v2end) {	// Catch bad data
//				System.err.println("  BAD: skipping");
				in.seek(id3v2end);
			    } else {
				b = new byte[size];
				in.read(b);
				if (tag.charAt(0)=='T' || tag=="COM" || tag=="PIC") {
				    String encoding = b[0]==0 ? "ISO-8859-1" : "UTF-16";
				    try {
					String val = new String(b, 1, b.length-1, encoding).trim();
                                        if (tag=="TAL") tag="TALB";
                                        else if (tag=="TCO") tag="TCON";
                                        else if (tag=="TP1") tag="TPE1";
                                        else if (tag=="TP2") tag="TPE2";
                                        else if (tag=="TCM") tag="TCOM";
                                        else if (tag=="COM") tag="COMM";
                                        else if (tag=="TT1") tag="TIT1";
                                        else if (tag=="TT2") tag="TIT2";
                                        else if (tag=="TT2") tag="TIT2";
                                        else if (tag=="TRK") tag="TRCK";
                                        else if (tag=="TYE") tag="TYER";
                                        else if (tag=="PIC") {
                                            setPictureTag(b, tags, true);
                                            tag=null;
                                        } else tag=null;
                                        if ("COMM".equals(tag) && val.indexOf("iTunes_CDDB_1")>=0) {
                                            tag="CDDB";
                                            val=val.substring(12);
                                        }
                                        if (tag!=null && (!tag.equals("COMM") || !badComment(val))) {
                                            tags.put(tag, val);
                                        }
				    } catch (Exception e) {}
				}
			    }
			}
		    } while (in.getFilePointer()<id3v2end);
		}
	    } else {
		in.seek(id3v2end);
	    }
	}

        in.seek(in.length()-128);
        if (in.read()=='T' && in.read()=='A' && in.read()=='G') {
            hasid3v1 = true;
        }
        in.seek(id3v2end);
        return id3version;
    }

    /**
     * If there is no Adler32 Checksum on the file, calculate it.
     */
    public void calculateChecksum() throws IOException {
        rfile.seek(id3v2end);
        Adler32 adler = new Adler32();
        byte[] buf = new byte[4096];
        int c;
        while ((c=rfile.read(buf, 0, buf.length))>0) {
            if (hasid3v1 && rfile.getFilePointer() > rfile.length()-128) {
                c -= rfile.getFilePointer() - (rfile.length()-128);
            }
            if (c>0) adler.update(buf, 0, c);
        }
        sum = adler.getValue();
    }

    private boolean badComment(String val) {
        if (val.indexOf("CDDB")>=0) return true;
        if (val.indexOf("YEAR: ")>=0) return true;
        if (val.indexOf("iTunNORM")>=0) return true;
        if (val.indexOf("iTunPGAP")>=0) return true;
        if (val.indexOf("Easy CD-DA")>=0) return true;
        try {
            Long.parseLong(val.substring(val.length()-8, val.length()), 16);
            return true;
        } catch (Exception e) {}
        try {
            Long.parseLong(val.substring(4, 12), 16);
            return true;
        } catch (Exception e) {}
        return false;
    }

    private boolean setPictureTag(byte[] b, HashMap tags, boolean id322) throws IOException {
        String encoding = b[0]==0 ? "ISO-8859-1" : "UTF-16";
        int i=4;
        if (!id322) for (i=1;b[i]!=0;i++);
        String mimetype = new String(b,1,i-1, "ISO-8859-1");
        if (!id322) i++;
        int pictype = b[i];
        if (pictype==0 || pictype==3 || pictype==112) {
            int oldi = i+1;
            if (encoding=="UTF-16") {
                for (i=oldi;b[i]!=0 || b[i+1]!=0;i+=2);
            } else {
                for (i=oldi;b[i]!=0;i++);
            }
            String description = new String(b,oldi,i-oldi, encoding);
            i += encoding=="UTF-16" ? 2 : 1;
            byte[] pic = new byte[b.length-i];

            if (mimetype.equals("PNG")) mimetype="image/png";
            else if (mimetype.equals("JPG")) mimetype="image/jpeg";
            else if (mimetype.equals("GIF")) mimetype="image/gif";
            if (pic.length>200 && mimetype.startsWith("image/")) {
                System.arraycopy(b,i,pic,0,pic.length);
                if (mimetype.equals("image/jpg")) mimetype="image/jpeg";
                tags.put("APIC.type", mimetype);
                if (description.length()>0) tags.put("APIC.description", description);
                tags.put("APIC.data", pic);
                return true;
            } else {
                return false;
            }
        } else {
            return false;
        }
    }

    /**
     * Read technical bits of MP3. Sets bitrate, samplerate, time, vbr
     */
    private void readBitrate(String file, RandomAccessFile in) throws IOException {
	final int[][] bitratetable = {
	    {0, 0, 0, 0, 0}, {32, 32, 32, 32, 8}, {64, 48, 40, 48, 16},
	    {96, 56, 48, 56, 24}, {128, 64, 56, 64, 32}, {160, 80, 64, 80, 40},
	    {192, 96, 80, 96, 48}, {224, 112, 96, 112, 56}, {256, 128, 112, 128, 64},
	    {288, 160, 128, 144, 80}, {320, 192, 160, 160, 96}, {352, 224, 192, 176, 112},
	    {384, 256, 224, 192, 128}, {416, 320, 256, 224, 144}, {448, 384, 320, 256, 160},
	    {0, 0, 0, 0, 0}
	};
	final int[][] sampleratetable = {
	    {44100, 22050, 11025}, {48000, 24000, 12000}, {32000, 16000, 8000}, {0, 0, 0}
	};
	final double[] vbrtable = {-1, 1152, 1152, 384 };

	byte[] head = new byte[4];
	long framestart;

	// Find first frame signature
	do {
	    framestart = in.getFilePointer();
	    in.read(head);
	    in.seek(++framestart);
	} while (framestart<in.length() && ((head[0]&0xFF)!=0xFF || (head[1]&0xE0)!=0xE0 || (head[1]&0x6)==0x00 || (head[2]&0xF0)==0xF0 || (head[2]&0x0C)==0x0C || (head[3]&0x03)==0x02));
	if (framestart>=in.length()) throw new IOException(file+": Can't find first frame");

	framestart--;
	int version = getBits(head[1], 3, 4);
	int layer = getBits(head[1], 1, 2);
	mode = getBits(head[3], 6, 7);

//	System.err.println(framestart+" - layer="+layer+" version="+version+" head="+Integer.toHexString(head[0]&0xff)+" "+Integer.toHexString(head[1]&0xff)+" "+Integer.toHexString(head[2]&0xff));

	bitrate = getBits(head[2], 4, 7);
	samplerate = getBits(head[2], 2, 3);
        boolean padding = getBits(head[2], 4, 4)==1;
	bitrate = bitratetable[bitrate][version==V1 ? 3-layer : layer==3 ? 3 : 4];
	samplerate = sampleratetable[samplerate][version==V1 ? 0 : version==V2 ? 1 : 2];

	// Look for Xing header which marks most (but not all, apparently) VBR encoded
	// 
	in.seek(framestart + (version==V1 ? (mode==MODE_MONO ? 21 : 36) : (mode==MODE_MONO ? 23 : 21)));
	in.read(head);
	vbr = (head[0]==88 && head[1]==105 && head[2]==110 && head[3]==103);

	if (vbr) {
	    int numBytes=(int)(in.length() - in.getFilePointer()), numFrames=-1;

	    in.read(head);
	    if ((head[3]&0x01)==0x01) numFrames = in.readInt();
	    if ((head[3]&0x02)==0x02) numBytes = in.readInt();

	    double tpf = (vbrtable[layer] / samplerate) / (version==V2 || version==V25 ? 2 : 1);
	    time = (int)(tpf*numFrames);
	    bitrate = (int) ((numBytes*8) / (tpf*numFrames*1000));
	} else {
            long bps = bitrate * 1000;
            long datasize = in.length()*8 - framestart;
            time = (int)(bps==0 ? 0 : datasize / bps);
        }
    }

    private static final int getBits(byte b, int s, int e) {
	return (b>>s) & ((1<<(e-s+1))-1);
    }

    public Map getTags()                { return tags; }
    public String getTag(String tag)    { return (String)tags.get(tag); }
    public boolean containsTag(String tag)    { return tags.containsKey(tag); }
    public String getTitle()		{ try{return getTag("TIT2");} catch(Exception e) {return "no title"; }}
    public String getArtist()		{ try{return getTag("TPE1"); } catch(Exception e) {return "no artist";}}
    public String getAlbum()		{ return getTag("TALB"); }
    //public String getYear()		{ return getTag("TYER"); }
    public int getYear() { return 0;}
    public String getComment()		{ return getTag("TCOM"); }
    public String getGenre()		{ try{return getTag("TCON"); } catch(Exception e) {return "Other";}}
    public int getTrackNo()		{ return getTag("TRCK")!=null ? Integer.parseInt(getTag("TRCK")) : 0; }
    public int getDiscNo()		{ return getTag("TPOS")!=null ? Integer.parseInt(getTag("TPOS")) : 0; }
    public int getDuration()		{ return time; }
    public long getSize()		{ return size; }
    public int getBitRate()		{ return bitrate; }
    public int getMode()		{ return mode; }
    public int getSampleRate()		{ return samplerate; }
    public float getGain()		{ return gain; }
    public String getOriginalGain()	{ return origgain; }
    public boolean isVBR()		{ return vbr; }
    public int getID3Version()          { return id3version; }  // eg 0x010000, 0x020300
    public long getChecksum()           { return sum; }
    public Date getDate()		{ return date; }        // Last Modified date of file
}
