/** This code is under the GPL, viewable at http://www.gnu.org/copyleft/gpl.html
 *  You are free to modify and redistribute the code as you see fit, but three main 
 *  conditions exist for usage of this code:
 *  1.  You may only pass on this code with the same rights and priveleges that were given
 *      to you (In other words, you must redistribute this license)
 *  2.  You may NOT distribute this code for sale or for profit.
 *  3.  There are no warranties about this code.  It is designed not to screw up your computer,
 *      but in the case that it does, the writer cannot be held responsible for losses.
 *
 *  Code author:  flug_m (post comments/questions to http://www.saveourtunes.com/forum/ In future
 *                versions, I'll make available alternative means of contact.)
 *  Code Revision date:  September 1, 2007
 *  Code Revision version: 0.5
 *
 *  This is an interface representing a song file on the hard drive. The goal is to present a
 *  clean and easy method of accessing field data from an MP3 or an AAC file.
 *
 *  Seeing as how existing code libraries for MP3s and AACs don't necessarily hide the work
 *  of reading the music file from the programmer, a wrapper class for the libraries will 
 *  need to bridge the available features of the library with the demands of this interface.
 *
 *  So far, the list of libraries to read ID3 data from an MP3 file is numerous.  I currently
 *  lean towards the javamusictag project, but would like to see if a better code library exists.
 *
 *  AAC libraries are almost nonexistent, considering that it is an Apple proprietary technology.
 *  I'd suggest looking into JHymn (or any aspect of the http://www.hymn-project.org/), but should
 *  the programmer have to resort to writing a library, many of the tags are easy to search for.
 *  Most fields are tagged as "com.apple.itunes.<fieldname>"; for example, to find the artist of a
 *  track, try looking for "com.apple.itunes.artist"; then, figure out the field size and extract 
 *  the data.
 *
 *
 *  Note:  When coding for AAC files under this interface, you are responsible for leaving
 *         existing DRM in place:  ourTunes is NOT intended to remove copy protection. You
 *         may only remove the following tags from audio files:
 *
 *         1. comments tag
 *         2. Apple ID (apID) (AAC files only)
 *
 */ 

package ourtunes.music;

public interface Song
{
	/**
	 *Get the year field from the MP3 or AAC song file.
	 *
	 *@Returns:
	 *   An integer indicating the year, or -1 if the field is not
	 *   present in the ID3 data.
	 */
	public int getYear();
	/**
	 *Get the artist field from MP3 or AAC song file.
	 *
	 *@Returns:
	 *   A String indicating the artist in the ID3 data, or 
	 *   an empty string if not present.
	 */
	public String getArtist();
	/**
	 *get the title, or name of the song, from the ID3 data in the 
	 *MP3 file (or equivalent field in AAC file).
	 *Don't return the file name, even if the Title
	 *of the song is missing from the ID3 data.
	 *
	 *@Returns:
	 *   A String indicating the title in the ID3 data, or 
	 *   an empty string if not present.
	 */
	public String getTitle();
	
	/**
	 *Get the genre of the song from MP3 or AAC song file.
	 *
	 *@Returns:
	 *    A String, or an empty string if the element is not
	 *    present in the header data.
	 */
	public String getGenre();
	
	/**
	 *Return a string representation of the class being used.
	 *For debugging purposes, output the file type (mp3 or aac),
	 *the artist name, and the title.  If any of the fields are missing,
	 *indicate that in this method.
	 *
	 *  @Returns: a string representation of the object.
	 */
	public String toString();
	
	//Note:  The following methods can be implemented as 
	//needed (just uncomment/add methods that provided in
	//future releases)  I don't require these methods to
	//be implemented because these features aren't often
	//filled in on songs.
	
	//public int getTrackNumber();
	//public String getAlbumName();
	//public String composer();
	
	/**
	 *Close any open stream.  If a class does not
	 *hold an open input stream on a file, then this
	 *method should do nothing.
	 */
	public void close();
	 
}