/** This code is under the GPL, viewable at http://www.gnu.org/copyleft/gpl.html
 *  You are free to modify and redistribute the code as you see fit, but three main 
 *  conditions exist for usage of this code:
 *  1.  You may only pass on this code with the same rights and priveleges that were given
 *      to you (In other words, you must redistribute this license)
 *  2.  You may NOT distribute this code for sale or for profit.
 *  3.  There are no warranties about this code.  It is designed not to screw up your computer,
 *      but in the case that it does, the writer cannot be held responsible for losses.
 *
 *  Code author:  flug_m (post comments/questions to http://www.saveourtunes.com/forum/ In future
 *                versions, I'll make available alternative means of contact.)
 *  Code Revision date:  September 1, 2007
 *  Code Revision version: 0.5
 *
 *  The HTTPHeader class contains methods that help identify and extract the header.
 *  This class relies on a few facts:
 *
 *  1.  Every HTTP header begins with a certain signature (see containsHeader).
 *  2.  The content-length field can be extracted with ease if present.  Note
 *      that this can be found in all Server response headers, but is worthless
 *      if we're not going to download the entire song
 *  3.  The end of the HTTP header is identified by two consecutive CRLF.
 *
 * Should there be a better method to identify the HTTP header, I'd encourage
 * development of this class.  As another note, I have to waste some space
 * by creating String objects and Pattern Matching those.  If there is an easier
 * way to do this (I understand that I can extract all this data without Strings, 
 * but I'm looking for ease of programming), then please develop this class.
 *
 * Note that all methods are static methods.  It is needless to create objects of this
 * class.
 */
 package ourtunes.net;
 

import java.util.regex.*;

public class HTTPHeader
{
	//Check if an HTTP header in present.  If so, return true.
	public static boolean containsHeader(byte[] data)
	{
		return containsHeader(new String(data));
	}
	public static boolean containsHeader(String data)
	{
		return (data.startsWith("GET")   || data.startsWith("POST")  ||
			    data.startsWith("HEAD")  || data.startsWith("HTTP")  || 
			    data.startsWith("TRACE") ||
			    data.startsWith("PUT")   || data.startsWith("DELETE")||
			    data.startsWith("CONNECT"));
	}
	
	/**Split the HTTP header.
	 *  
	 * This method looks for two consecutive CRLFs, and returns the index
	 * of the beginning of the data (excluding the CRLF)s.
	 * Should this index not be found, the method will return 0.
	 */
	public static int splitHeader(byte[] data)
	{
		byte carraigeReturn = 13;
		byte lineFeed = 10;
		
		for (int i=0; i<data.length-3; i++)
		{
			if (data[i] == carraigeReturn)
			{
				if (data[i+1] == lineFeed && data[i+2] == carraigeReturn &&
					data[i+3] == lineFeed) return (i+4);
			}
		}
		return 0;
	}
	
	/**
	 * Extract the content length from the HTTP header.
	 * This method uses Pattern Matching to identify the 
	 * field and extract its value.
	 *
	 * As a note, content-lengths can be quite lengthy, but
	 * an signed integer can represent a file size of 0 bytes - 2GB.
	 * If you can identify an MP3 larger than 2 GB, then the method
	 * needs to return a long.
	 *
	 * @Returns:  an int indicating the length.
	 */
	public static int getContentLength(byte[] data)
	{
		String str = new String(data);
		Pattern p = Pattern.compile("content-length: ([0-9]*)", Pattern.CASE_INSENSITIVE);
		Matcher m = p.matcher(str);
		
		if (m.find())
			return Integer.parseInt(m.group(1));
		else return 0;
	}
}