/** This code is under the GPL, viewable at http://www.gnu.org/copyleft/gpl.html
 *  You are free to modify and redistribute the code as you see fit, but three main 
 *  conditions exist for usage of this code:
 *  1.  You may only pass on this code with the same rights and priveleges that were given
 *      to you (In other words, you must redistribute this license)
 *  2.  You may NOT distribute this code for sale or for profit.
 *  3.  There are no warranties about this code.  It is designed not to screw up your computer,
 *      but in the case that it does, the writer cannot be held responsible for losses.
 *
 *  Code author:  flug_m (post comments/questions to http://www.saveourtunes.com/forum/ In future
 *                versions, I'll make available alternative means of contact.)
 *  Code Revision date:  March 21, 2008
 *  Code Revision version: 0.7
 *
 *  The NetTrafficListener is the thread actually responsible for saving music files.
 *  Its inner class is the TrafficListener for use on the Pass-through thread; this
 *  class isn't easy to instantiate, so a convenience method is provided to return
 *  a listener for the class (In a future version, it should probably return
 *  one listener per class, not a new one per each call).
 *
 *  Should an HTTP header that indicates a music transfer be detected, a file
 *  is created for storing the data, and then it is later renamed (when filled)
 *  to make it usable.
 *
 *  This class is probably the messiest of them all, and the least explained. Apologies
 *  in advance.
 *
 *  Changes 
 *  8/30/07  Threads report to gui.  Better for user.
 *  8/31/07  temp files are created in the default directory, and saved into the user specified directory.
 *  8/31/07  temp files will delete if the aren't completed.  As a minor bug, this is not true if
 *           the thread does not properly die.
 *  11/28/07 Bug fix:  This thread was being blocked with a content length of 0 left to complete a file.
 *           This in turn prevented the thread from completing execution; this is now avoided.
 *  1/23/08  The TrafficListener was storing both passing client AND server data.  However,
 *           the music comes purely off the client connection and while the server connection is mostly 
 *           inactive, it could potentially screw up the file.   After this fix, the byte data is identified,
 *           and client data is always discarded.
 *   2/22/08 Use the locking mechanism on the GUI status panel instead of relying on synchronization.  See
 *           Change 2.22.08 ourtunes.gui.OurTunesPanel
 *   3/21/08 Removed obsolete method detectInterrupt().  This was corrected with 1.23.08 where the data
 *           ID was recorded into the bytearray.
 *   7/23/08 Support added for midi and wav files; they are now recognized by the filters.
 */
 package ourtunes.net;
 
 
import java.net.*;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.Scanner;
import java.util.regex.*;
import java.io.*;
import javax.swing.JOptionPane;

import ourtunes.Shared;
import ourtunes.Logger;

public class NetTrafficListener extends Thread
{
	public class Trafficker implements TrafficListener
	{
		public void trafficReceived(TrafficEvent event)
		{
			//just in case we received an interrupt, my method
			//of stopping the file download is to place a three-byte
			//array into the queue with values {3,1,0}.  I do imagine
			//that this should not coincide with real data.
			//Otherwise,		
			//place the item into the queue; we don't expect it to be full.
			try{
			if (event.getOrigin() == TrafficEvent.INTERRUPT_EVENT)
			{
				byte[] arr = {3,1,0};
				queue.put(new ByteArray(arr, event.getOrigin()));
				event.interruptLaunchingThread();
			}
			else //queue.put(event.getByteArray().setID(event.getOrigin()));
			{
				ByteArray b = event.getByteArray();
				b.setID(event.getOrigin());
				queue.put(b);
			}
			} catch (Exception e) {}
		}
	}
	public NetTrafficListener(Thread owner)
	{
		super("NetTrafficListener("+owner.getName()+")");
		queue = new LinkedBlockingQueue<ByteArray>(200);
		this.owner = owner;
		logger = Logger.getLogger();
	}
	//Return a TrafficListener (which is an inner class) that can access the elements of this Thread.
	public Trafficker getTrafficker()
	{
		return new Trafficker();
	}
	public void run()
	{
		do
		{
		//obtain the event in the queue.
		try{
		ByteArray ba = queue.take();
		byte[] data = ba.getBytes();
		
		//test the array to detect an interrupt sequence.
		//purge the queue elements (because its contents would no longer be useful in this thread),
		//and return (an interrupt indicates that the pass-through thread stopped and that no more events
		//are being generated).
		if (ba.getID() == TrafficEvent.INTERRUPT_EVENT)
		{queue.clear(); logger.println("received an interrupt... ending thread."); return;}
		
		String sample = null;
		//If we're not requesting music data, then pass over it.
		if (HTTPHeader.containsHeader(data))
        {
            int datal = HTTPHeader.splitHeader(data);
            sample = new String(data, 0, datal-2);
            logger.println(sample);
        }
        else sample = "";
		Pattern p = Pattern.compile("GET .*(mp3|m4a|midi|wav)", Pattern.CASE_INSENSITIVE);
		Matcher match = p.matcher(sample);

		if (match.find()) 
		{
		String ext = match.group(1);
		fillMusic(ext, data);
		return;
		}
		
		if(sample.indexOf("logout") != -1)
		{
		  //Need to logout of the server... Don't keep the socket open
		  Shared.panel.setStatus("the client unexpectedly logged out from the server.  Try again later.");
		  owner.interrupt();
		  return;
		}
		} catch(Exception e) {e.printStackTrace();}
		} while (true); //loop this thread indefinitely
	}
	//This method actually handles the data copy from the queue.
	//Fix - 8/30/08.  I am skipping several bytes worth of header data by leaving out the first packet!!!
	private void fillMusic(String extension, byte[] data)
	{
		boolean hasLock = Shared.panel.isStatusAvailable();
		if (hasLock) {Shared.panel.lock(); logger.println("Lock Acquired!");}
		
		//generate the File name.  We'll use a random name by default, and later 
		//when the data is filled (and the ID3 tags can be read), we can change 
		//the file name to user requested file name.
		String fileName = new String (Shared.getRandom()+".bin."+extension);
		File f = new File(Shared.getProperty("home"), fileName);
		try{
		FileOutputStream fout = new FileOutputStream(f);
		
		int length = -1;
		int dataStart = -1;
		//Need to know how many bytes to take in the stream.
		ByteArray array = queue.take();
		data = array.getBytes();
		if (HTTPHeader.containsHeader(data))
        {
            dataStart = HTTPHeader.splitHeader(data);
            String sample = new String(data, 0, dataStart-2);
            logger.println(sample);
        }
        
		if (HTTPHeader.containsHeader(data))
		{
			length = HTTPHeader.getContentLength(data);
			dataStart = HTTPHeader.splitHeader(data);
		}
		else
		{
			//There's a problem:  we didn't read a header to get the data length
			//we'll just return creating a 0-byte length file.
			Shared.panel.setStatus("No header data.  Download will not complete.");
			logger.println("no HTTP header - returning with 0 byte length file..");
			fout.close();
			f.delete();
			return;
		}
		data = array.getArrayStartingAt(dataStart);
		
		//"priming" the loop
		fout.write(data);
		length -= data.length;
		Shared.panel.appendDownloadStatus(length, extension);
		
		while (length > 0)
		{
			array = queue.take();
			if (array.getID() == TrafficEvent.INTERRUPT_EVENT)
			{
				logger.println("Thread interrupted. Download not completed.");
				if (hasLock)   {hasLock=false; Shared.panel.unlock();}
				f.delete();
				return;
			}
			if (array.getID() != TrafficEvent.FROM_SERVER) continue;
			//If we exceed the Content-Length value, then we may need 
			//to find the end of the byte range.
			if (array.getLength() > length)
				data = array.getRange(0,length);
			else data = array.getBytes();
			
			//Lock the GUIpanel by this thread.
			if (!hasLock)
			{
				hasLock = Shared.panel.isStatusAvailable();
				if (hasLock) {Shared.panel.lock();
				logger.println("Lock Acquired!");}
			}
			
			fout.write(data);
			length -= data.length;
			Shared.panel.appendDownloadStatus(length, extension);
			logger.println("num bytes left:  "+length+"\tnum read:  "+data.length);
			//fill in the array
		}
		
		//The file is written, let's pass it off to someone else to appropriately name it.
		//Interrupt this thread to stop its streaming abilities.
		logger.print("----------------------------------\n");
		logger.print("-----Completed download.");
		logger.println("");
		owner.interrupt();
		if (owner instanceof PassThrough)
		{PassThrough pt = (PassThrough) owner;
			pt.closeSockets();}
		fout.close();
		Shared.fixFile(f, extension);
		if (hasLock) {hasLock = false; Shared.panel.unlock();}
		}		
		//Catching the exception from output is easy, what to do with it, I have no clue.
		catch(Exception e) {
			JOptionPane.showMessageDialog(null, e.toString()+"\nThis program will terminate shortly."+
				" To view errors with ourTunes, launch the jar file from the Console or Terminal.");
			e.printStackTrace();
			System.exit(1);
		}
	}
	
	private BlockingQueue<ByteArray> queue;
	private Thread owner;
	private Logger logger;
}