/** This code is under the GPL, viewable at http://www.gnu.org/copyleft/gpl.html
 *  You are free to modify and redistribute the code as you see fit, but three main 
 *  conditions exist for usage of this code:
 *  1.  You may only pass on this code with the same rights and priveleges that were given
 *      to you (In other words, you must redistribute this license)
 *  2.  You may NOT distribute this code for sale or for profit.
 *  3.  There are no warranties about this code.  It is designed not to screw up your computer,
 *      but in the case that it does, the writer cannot be held responsible for losses.
 *
 *  Code author:  flug_m (post comments/questions to http://www.saveourtunes.com/forum/ In future
 *                versions, I'll make available alternative means of contact.)
 *  Code Revision date:  February 22, 2008
 *  Code Revision version: 0.7
 *
 *  The pass-through thread bears the majority of the work of the program, and
 *  revisions should be minimized.
 *
 *  Once the host thread accepts a new connection, it launches a new
 *  pass through thread that handles the data transfer.
 *
 *  The thread first checks the SocketAddress that it connects to.  To prevent
 *  infinite loops, the SocketAddress is either routed to a valid server
 *  on the network, or to iTunes through the loopback address.
 *
 *  Following the check on the SocketAddress, it will perform the pass-through.
 *  If no activity passes through the thread over 10 seconds, it is allowed to 
 *  expire.
 *
 *  It is important that this thread have the ability to be interrupted at any time.
 *  Unfortunately, this doesn't always happen, and development contributions would
 *  be greatly welcomed.
 *
 *  object names are generally explained, or are generally a shortened name (ie, I'll commonly
 *  use the object name "f" for Files).
 *
 *  Final notes:  It seems that iTunes can only buffer so much data at a time.  Sending
 *        more data than it can handle will cause the iTunes socket to overflow, iTunes
 *        will close the port and disassociate from the shared library.  This interrupts
 *        our file saves, so I log iTunes traffic to file to meter the pass-through.
 *
 *        Anytime the JOptionPane is used, it prevents all other GUI elements from
 *        gaining focus.  This is a problem for the user;  I provided a simple
 *        Message class that behaves in the same way as a JOptionPane.showMessageDialog(...)
 *        call would, but I haven't put it to use. 
 *
 * Changes:
 *  8/30/07  Threads can report to the status pane and to console.  Better for user.
 *  9/06/07  Instead of having to resolve a service instance name, The PassThrough receives
 *           an InetAddress that is already resolved.
 * 11/19/07  Passthrough threads won't die as long as the client connection is open.
 * 11/20/07  Change in behavior.  Threads looping under the conditions below will sleep more often:
 *			  a) Thread has no activity for >10 seconds;
 *            b) Client socket is open.
 * 11/26/07  Removed extraneous code that kept the thread suspended when no client data came through.
 * 11/26/07  ourTunes usage does not log to file.
 *  2/22/08  PassThrough.java is affected by the changes to the locking mechanism;  See 2.22.08
 *           ourtunes.gui.OurTunesPanel for a better explanation.
 */
 package ourtunes.net;
 
import java.io.*;
import java.net.*;
import javax.swing.JOptionPane;
import java.util.List;

import ourtunes.net.*;
import ourtunes.Shared;
import ourtunes.Logger;

class PassThrough extends Thread
{
	public PassThrough(Socket s)
	{
		super("Pass-through-" + counter++);
		iTunesClient = s;
	    logger = Logger.getLogger();
		
		//Set up the alternate thread for downloading the music.
		NetTrafficListener ntl = new NetTrafficListener(this);
		trafficl = ntl.getTrafficker();
		ntl.start();
		recordData = true;
		isInterrupted = false;
		log = null;
		
	}
	
	public void run()
	{
	    noDelay = Shared.getProperty("port.FirstOpen").equals("1");
	    if (noDelay) Shared.setProperty("port.FirstOpen", "0");
	    
		boolean error = false;
		long oldTime = System.currentTimeMillis();
		server = Shared.getServer();
		TrafficEvent event;
		try{
		//Check the request to see who originated it.  If the request came
		//from this computer, then route the request to a host; if the 
		//request came from a networked computer, then route it to iTunes (through
		//the loopback interface).
		InetSocketAddress dr= (InetSocketAddress) iTunesClient.getRemoteSocketAddress();
		if (!Shared.isHostAddress(dr.getAddress()) || server == null)
		{iTunesServer = new Socket("127.0.0.1", 3689); recordData = false;
		 Shared.panel.setStatus(
		 	"<code style=\"color:red\">A network user connected to your library</code>");
		}
		else 
		//Validate the host.  If the host address is an InetAddress on this computer,
		//Then use the loopback address (to avoid an infinite loop with the program constantly
		//sending out data to itself).
		if (Shared.isHostAddress(server))
		{iTunesServer = new Socket("127.0.0.1", 3689); recordData = false;
		Shared.panel.setStatus("<code style=\"color:green\">You selected your library.  This "+
			"will not pass.<code>");
		}
		else iTunesServer = new Socket(server, 3689);	
		
		
		//Set up all the input/output streams to record the data.
		//log = new File("./log.txt");
		//fout = new PrintStream(log);
		cis = iTunesClient.getInputStream();
		cos = iTunesClient.getOutputStream();
		sis = iTunesServer.getInputStream();
		sos = iTunesServer.getOutputStream();
		} catch (Exception e) {e.printStackTrace(); System.exit(1);}
		
		try{
		do{
		//perform the actual pass-through	
		while (cis.available() > 0)
		{
			b = new byte[cis.available()];
			cis.read(b);
			sos.write(b);
			sos.flush();
			event = new TrafficEvent(b, TrafficEvent.FROM_CLIENT, this);
			logger.println("writing data to server...");
			oldTime = System.currentTimeMillis();
			
			//Alert the listener to traffic received.
			if (recordData)
			trafficl.trafficReceived(event);
		}
		
		while (sis.available() > 0)
		{
			b = new byte[sis.available()];
			sis.read(b);
			
			try{
				if (!error) cos.write(b);
				
			} catch(Exception f) {
				if (!error)
				{
					if (Shared.panel.isStatusAvailable())
				Shared.panel.setStatus("<p style=\"color:green\">The socket to the client iTunes "+
				"closed.  To terminate the download, click interrupt pass-through in Options.</p>");
				error = true;}
			}
			event = new TrafficEvent(b, TrafficEvent.FROM_SERVER, this);
			oldTime = System.currentTimeMillis();
			//fout.write(b);
			
			//Alert the listener to traffic received.
			if (recordData)
			trafficl.trafficReceived(event);
		}
		//Interrupt mechanism for the thread.
		//I chose to throw an exception instead of break the loop.
		if (this.isInterrupted() || isInterrupted)
		{
			logger.println("Interrupted due to completed download.");
			error = true;
			throw new InterruptedException("Download complete");
		}
		
		if (System.currentTimeMillis() - oldTime > 10000 && iTunesClient.isConnected())
			{logger.println("went to sleep"); sleep(9990); }//9.99 seconds
		
		//Allow the thread to timeout after 10 seconds of inactivity. 
		} while ((System.currentTimeMillis()-oldTime) < 10000 || (iTunesClient.isConnected() && iTunesServer.isConnected()));
		}catch (InterruptedException ie){
			ie.printStackTrace();
		}catch (Exception e) {e.printStackTrace(); 
			
			//Report to the gui
			if (Shared.panel.isStatusAvailable())
			Shared.panel.setStatus("<p style=\"color:#FF8C00\">The socket to the iTunes "+
				"client closed.  See console/terminal for more details.</p>");		
		}
		
		trafficl.trafficReceived(new TrafficEvent(null, TrafficEvent.INTERRUPT_EVENT, this));
		logger.println("end of a thread run() method!");
	}
	protected void finalize()
	{
		System.out.println("Pass-through thread death - Garbage Collection!");
		try{
		iTunesClient.close();
		iTunesServer.close();
		//fout.close();
		//log.delete();
		
		} catch (Exception e) {e.printStackTrace();}
	}
	
	public void interrupt()
	{
		super.interrupt();
		isInterrupted = true;
		
	}
	public boolean isInterrupted()
	{
		if (isInterrupted) logger.println("Pass-through thread was interrupted");
		return (super.isInterrupted() || isInterrupted);
	}
	public void closeSockets()
	{
		finalize();
	}
	
		private Socket iTunesClient;
		private Socket iTunesServer;
		private InputStream cis;    //Client InputStream
		private OutputStream cos;   //Client OutputStream
		private InputStream sis;   //Server InputStream
		private OutputStream sos;  //Server OutputStream
		private PrintStream fout;  //File OUTputstream
		private byte[] b=null;     //caching array
		private TrafficListener trafficl;  //thread specific traffic listener.
		private InetAddress server;
		private File log;
		private boolean recordData;
		private boolean isInterrupted;
		private boolean noDelay;
		static int counter = 1;
		private Logger logger;
}